package com.jetbyte.cache5;

import java.lang.ref.SoftReference;
import java.lang.ref.ReferenceQueue;

public interface Cache
{
   public Object find(String keyString);
   public Object find(Key cacheKey);

   public void add(String keyValue, Object obj);
   public void add(Key cacheKey, Object obj);
   
   public void remove(String keyValue);
   public void remove(Key cacheKey);
   
   public String name();
   public String type();

   public int entries();
   public int hits();
   public int misses();
   
   public void flush();

   public class Key implements Comparable
   {
      public Key()
      {
         setKey("");
      }

      public Key(String key)
      {
         keyValue = key;
         hashCode = keyValue.hashCode();
      }

      public void setKey(String key)
      {
         keyValue = key;
         hashCode = keyValue.hashCode();
      }

      public String toString()
      {
         return keyValue;
      }

      public int hashCode()
      {
         return hashCode;
      }

      public boolean equals(Object obj)
      {
         if (this == obj)
         {
            return true;
         }
         else if (obj == null || getClass() != obj.getClass())
         {
            return false;
         }

         Key rhs = (Key) obj;

         return keyValue.equals(rhs.keyValue);
      }

      public int compareTo(Object obj)
      {
         if (this.equals(obj))
         {
            return 0;
         }

         Key rhs = (Key) obj;

         return keyValue.compareTo(rhs.keyValue);
      }

      private String keyValue;

      private int hashCode;
   }

   public static class ObjectRef extends SoftReference
   {
      protected ObjectRef(
         Cache cache, 
         Key key, 
         Object obj)
      {
         super(obj,collectedQueue);
         this.hardRef = obj;
         this.key = key;
         this.cache = cache;

         lastUsedTime = System.currentTimeMillis(); 
      }

      synchronized public Object getObject()
      {
         lastUsedTime = System.currentTimeMillis(); 

         hardRef = get();        // if the object has been released, we have a chance to 
                                 // make it valid again...   

         return hardRef;
      }

      protected String getKey()
      {
         return key.toString();
      }

      synchronized protected long getLastUsedTime()
      {
         return lastUsedTime;
      }

      synchronized protected void releaseObject()
      {
         hardRef = null;
      }

      protected void removeFromCache()
      {
         cache.remove(key);
      }

      protected void destroy()
      {
         cache = null;
         key = null;
         hardRef = null;
      }

      private Cache cache;
      private Key key;
      private Object hardRef;
      private long lastUsedTime = 0;

      private static ReferenceQueue collectedQueue = new ReferenceQueue();

      private static GarbageMonitor monitor = GarbageMonitor.create(collectedQueue);

      private static class GarbageMonitor extends Thread
      {
         private GarbageMonitor(ReferenceQueue queue)
         {
            this.queue = queue;
         }

         public void run()
         {
            while (true)
            {
               try
               {
                  ObjectRef ref = (ObjectRef)queue.remove();

                  System.out.println("Garbage collected: " + ref.getKey());

                  ref.removeFromCache();

                  ref.destroy();
               }
               catch(InterruptedException e)
               {

               }
               catch(Exception e)
               {
                  System.out.println("Exception caught");
                  break;
               }
            }
         }

         private static GarbageMonitor create(ReferenceQueue queue)
         {
            GarbageMonitor mon = new GarbageMonitor(queue);

            mon.setDaemon(true);

            mon.start();

            return mon;
         }

         private ReferenceQueue queue;
      }
   }
}
